from datetime import datetime
from typing import (
    Any,
    Callable,
    Iterator,
    Literal,
    TypeAlias,
    TypedDict,
    TypeVar,
    cast,
    overload,
)

from splatnet3_scraper.query.json_parser import JSONParser
from splatnet3_scraper.utils import match_partial_path

T = TypeVar("T")
S = TypeVar("S")

PathType: TypeAlias = str | int | tuple[str | int, ...]


class MetaData(TypedDict, total=False):
    query: str
    timestamp: float


class QueryResponse:
    """The QueryResponse class represents a response from the SplatNet 3 API.
    The class provides various convenience methods for interacting with the
    returned data and contains metadata about the response. The metadata
    contains the query that was used to get the data and the timestamp of the
    response. Additional metadata may be added in the future. Using the getitem
    method on the response will return another QueryResponse with the data from
    the key. If the data is a list, the getitem method will generate a
    QueryResponse for each item in the list. The QueryResponse class also
    provides a parse_json method that returns a JSONParser object for the
    response data. This is currently not used by the library, so it may be
    removed in the future.
    """

    def __init__(
        self,
        data: dict[str, Any] | list[dict[str, Any]],
        metadata: dict[str, Any] | MetaData | None = None,
    ) -> None:
        """Initializes a QueryResponse.

        Args:
            data (dict[str, Any]): The data from the response as a dictionary.
                If the data is a list, the data should be a list of
                dictionaries. The data should be some sort of JSON object of
                nested dictionaries and lists.
            metadata (dict[str, Any] | None): The metadata from the response.
                Possible keys are "query" and "timestamp". Defaults to None.
        """
        self._data = data

        self._metadata = self.__parse_metadata(metadata)

    def __parse_metadata(
        self, metadata: dict[str, Any] | MetaData | None
    ) -> MetaData | None:
        """Parses the metadata from the response.

        Args:
            metadata (dict[str, Any] | MetaData | None): The metadata from the
                response.

        Returns:
            MetaData | None: The parsed metadata.
        """
        if metadata is None:
            return None

        annotations = MetaData.__annotations__
        _metadata = {}
        for key, value in metadata.items():
            if key in annotations:
                _metadata[key] = value
        if len(_metadata) == 0:
            return None
        return MetaData(_metadata)  # type: ignore

    def __top_level_is_list(self) -> bool:
        """Returns whether the top level of the data is a list. Internal use
        only.

        Returns:
            bool: Whether the top level is a list.
        """
        return isinstance(self._data, list)

    @property
    def data(self) -> dict[str, Any] | list[dict[str, Any]]:
        """The raw data from the response. This is the data that was passed to
        the QueryResponse at initialization.

        This property, as opposed to the ``__getitem__`` method, will return the
        raw data from the response without any parsing.

        Returns:
            dict[str, Any] | list[dict[str, Any]]: The raw data.
        """
        return self._data

    @property
    def metadata(self) -> MetaData:
        """The metadata from the response.

        Raises:
            ValueError: If no metadata was provided at initialization.

        Returns:
            MetaData: The metadata.
        """
        if self._metadata is None:
            raise ValueError("No metadata was provided.")
        return self._metadata

    @property
    def query(self) -> str:
        """The query that was used to get the data.

        Raises:
            ValueError: If no metadata was provided at initialization.

        Returns:
            str: The query.
        """
        if self._metadata is None:
            raise ValueError("No metadata was provided.")
        try:
            return self._metadata["query"]
        except KeyError:
            raise ValueError("No query was provided.")

    @property
    def timestamp_raw(self) -> float:
        """The timestamp of the response, as a float.

        Raises:
            ValueError: If no metadata was provided at initialization.

        Returns:
            float: The timestamp.
        """
        if self._metadata is None:
            raise ValueError("No metadata was provided.")
        try:
            return self._metadata["timestamp"]
        except KeyError:
            raise ValueError("No timestamp was provided.")

    @property
    def timestamp(self) -> datetime:
        """The timestamp of the response, as a datetime object.

        Returns:
            datetime: The timestamp.
        """
        return datetime.fromtimestamp(self.timestamp_raw)

    def parse_json(self) -> JSONParser:
        """The JSONParser object containing the data.

        Returns:
            JSONParser: The data.
        """
        return JSONParser(self._data)

    def __repr__(self) -> str:
        """Returns a string representation of the QueryResponse.

        Representation includes the metadata, if it exists. If the metadata
        exists, the representation will include the query and the timestamp.
        This method also iterates over the MetaData dictionary and adds it to
        the representation. This method should not need updating if new
        metadata fields are added.

        Returns:
            str: The string representation of the QueryResponse.
        """
        out_str = "QueryResponse("
        if self._metadata is None:
            return out_str + ")"

        spacing = " " * len(out_str)
        for key, value in self._metadata.items():
            # Intercept if key is "timestamp" and convert to datetime
            if key == "timestamp":
                value = cast(float, value)
                out_value = datetime.fromtimestamp(value).strftime(
                    "%Y-%m-%d %H:%M:%S"
                )
            elif isinstance(value, str):
                out_value = value if len(value) < 20 else value[:20] + "..."
            elif isinstance(value, float):
                out_value = f"{value:.2f}"
            else:
                out_value = str(value)

            out_str += spacing + f"{key}={out_value},\n"

        return out_str[:-2] + ")"

    def __eq__(self, other: object) -> bool:
        """Returns whether the QueryResponse is equal to another object.

        Explicitly, this method checks whether the other object is a
        QueryResponse and whether the data and metadata are the same as the ones
        found in the QueryResponse.

        Args:
            other (object): The object to compare to.

        Returns:
            bool: Whether the QueryResponse is equal to the other object.
        """
        if not isinstance(other, QueryResponse):
            return False
        return self._data == other._data and self._metadata == other._metadata

    def __len__(self) -> int:
        return len(self._data)

    def __getitem__(self, key: PathType) -> "QueryResponse":
        """Returns a QueryResponse object containing the data at the given
        key. If the key is a tuple, this method will treat it as taking multiple
        keys in order to get to the data. For example, the following two are
        equivalent:

        >>> query_response["key1"]["key2"] == query_response["key1", "key2"]

        This method will then return a QueryResponse object containing the data
        at the given key.

        Args:
            key (PathType): The key to get the data

        Returns:
            QueryResponse: The QueryResponse object containing the data.
        """
        if isinstance(key, tuple):
            for k in key:
                self = self[k]
            return self

        data = self._data[key]  # type: ignore
        if isinstance(data, (dict, list)):
            return QueryResponse(data, metadata=self._metadata)
        return data

    def keys(self) -> list[str | int]:
        """Returns a list of keys in the data. If the top level of the data is
        a list, this method will return a list of integers from ``0`` to the
        length of the list.

        Returns:
            list[str | int]: The keys in the data.
        """
        if self.__top_level_is_list():
            return list(range(len(self)))
        data = cast(dict[str, Any], self._data)
        return list(data.keys())

    def values(self) -> list[Any]:
        """Returns a list of values in the data.

        Returns:
            list[Any]: The values in the data.
        """
        return list(self)

    def items(self) -> list[tuple[str, Any]]:
        """Returns a list of items in the data. If the top level of the data is
        a list, this method will return a list of integers from ``0`` to the
        length as the keys.

        Done to match the behavior of ``dict.items()``, which returns a list of
        tuples of the form ``(key, value)``. This method will return a list of
        tuples of the form ``(key, value)`` where the key can be an integer if
        the top level of the data is a list, otherwise it works the same as
        ``dict.items()``.

        Returns:
            list[tuple[str, Any]]: The items in the data.
        """
        if self.__top_level_is_list():
            return list(enumerate(self))  # type: ignore
        data = cast(dict[str, Any], self._data)
        return list(data.items())

    def __iter__(self) -> Iterator[Any]:
        """Returns an iterator over the values in the data.

        This method simply calls ``self.keys()`` and iterates over over
        ``self[key]`` for each key in the list of keys. This method is used to
        implement ``self.values()``.

        Yields:
            Iterator[Any]: The values in the data.
        """
        keys = self.keys()
        for key in keys:
            yield self[key]

    @overload
    def show(self, return_value: Literal[False]) -> None:
        ...

    @overload
    def show(self, return_value: Literal[True]) -> dict[str, Any]:
        ...

    def show(
        self, return_value: bool = False
    ) -> dict[str, Any] | list[dict[str, Any]] | None:
        """Prints the data to the console. If ``return_value`` is True, returns
        the data as a dict instead.

        Args:
            return_value (bool): Whether to return the data as a dict. If True,
                the data will be returned instead of printed. Defaults to False.

        Returns:
            dict[str, Any] | list[dict[str, Any]] | None: The data. If
                return_value is False, this will be None.
        """
        if return_value:
            return self._data
        print(self._data)
        return None

    def apply(
        self,
        func: Callable[[Any], T],
        key: PathType | list[PathType],
        partial: bool = True,
    ) -> T | list[T]:
        """Applies a function to the data.

        Given a function and a key to apply the function to, this method will
        apply the function to the data at the given key. If the key is a tuple,
        this method will treat it as a path. For example, a ``key`` argument of
        ``(0, "key1")`` will be treated as ``data[0]["key1"]``. If the key is
        a string, this method will treat it as a key in a dictionary. Integers
        will be treated as indices in a list. If the ``partial`` argument is
        ``True``, this method will apply the function to all keys in the data
        that match the given key or path. For example, if the ``key`` argument
        is ``(0, "key1")`` and the ``partial`` argument is ``True``, this will
        apply the function to all values within the JSON object where the path
        is ``...[0]["key1"]``. If the ``partial`` argument is ``False``, this
        will only apply the function to the value at the absolute key or path in
        the data.

        Args:
            func (Callable[[Any], T]): The function to apply to the data. The
                function must take a single argument and return a value of type
                ``T``. The argument that the function takes will be
                representative of the value at the given key or path. If the
                ``partial`` argument is ``True``, the argument will be a
                representative of the value at the given key or path within the
                data.
            key (PathType | list[PathType]): The key or path to apply the
                function to. If the key is a tuple, this method will treat it
                as a path. For example, a ``key`` argument of ``(0, "key1")``
                will be treated as ``data[0]["key1"]``. If the key is a string,
                this method will treat it as a key in a dictionary. Integers
                will be treated as indices in a list. If the ``partial``
                argument is ``True``, this method will treat a ``key``
                argument of ``(0, "key1")`` as ``...[0]["key1"]`` rather than
                just ``data[0]["key1"]``.
            partial (bool): Whether the given key or path is a partial path. If
                ``True``, this method will treat a ``key`` argument of ``(0,
                "key1")`` as ``...[0]["key1"]`` rather than just
                ``data[0]["key1"]``. If ``False``, this method will treat a
                ``key`` argument of ``(0, "key1")`` as ``data[0]["key1"]``.
                Defaults to ``True``.

        Returns:
            T | list[T]: The result of applying the function to the data. If
                multiple paths match the given key or path, this will be a list
                of the results of applying the function to each path. If only a
                single path matches the given key or path, this will be the
                result of applying the function to that path.
        """
        if not partial:
            if not isinstance(key, list):
                return func(self[key])
            out: list[T] = []
            for path in key:
                out.append(func(self[path]))

        paths = self.match_partial_path(key)
        return [func(self[path]) for path in paths]

    def apply_reduce(
        self,
        func: Callable[[Any], T],
        reduce_func: Callable[[list[T]], S],
        key: PathType | list[PathType],
        partial: bool = True,
    ) -> S:
        """Applies a function to the data and then reduces the result.

        Given a function, a reduce function, and a key to apply the function to,
        this method will apply the function to the data at the given key. If the
        key is a tuple, this method will treat it as a path. For example, a
        ``key`` argument of ``(0, "key1")`` will be treated as
        ``data[0]["key1"]``. If the key is a string, this method will treat it
        as a key in a dictionary. Integers will be treated as indices in a list.
        If the ``partial`` argument is ``True``, this method will apply the
        function to all keys in the data that match the given key or path. For
        example, if the ``key`` argument is ``(0, "key1")`` and the ``partial``
        argument is ``True``, this will apply the function to all values within
        the JSON object where the path is ``...[0]["key1"]``. If the ``partial``
        argument is ``False``, this will only apply the function to the value at
        the absolute key or path in the data.

        Args:
            func (Callable[[Any], T]): The function to apply to the data. The
                function must take a single argument and return a value of type
                ``T``. The argument that the function takes will be
                representative of the value at the given key or path. If the
                ``partial`` argument is ``True``, the argument will be a
                representative of the value at the given key or path within the
                data.
            reduce_func (Callable[[T], S]): The function to reduce the result
                of applying the function to the data. The function must take a
                single argument and return a value of type ``S``. The argument
                that the function takes will be the result of applying the
                function to the data.
            key (PathType | list[PathType]): The key or path to apply the
                function to. If the key is a tuple, this method will treat it
                as a path. For example, a ``key`` argument of ``(0, "key1")``
                will be treated as ``data[0]["key1"]``. If the key is a string,
                this method will treat it as a key in a dictionary. Integers
                will be treated as indices in a list. If the ``partial``
                argument is ``True``, this method will treat a ``key``
                argument of ``(0, "key1")`` as ``...[0]["key1"]`` rather than
                just ``data[0]["key1"]``.
            partial (bool): Whether the given key or path is a partial path. If
                ``True``, this method will treat a ``key`` argument of ``(0,
                "key1")`` as ``...[0]["key1"]`` rather than just
                ``data[0]["key1"]``. If ``False``, this method will treat a
                ``key`` argument of ``(0, "key1")`` as ``data[0]["key1"]``.
                Defaults to ``True``.

        Returns:
            S: The result of applying the function to the data and then reducing
                the result. This will always be a single value of type ``S``.
        """
        out = self.apply(func, key, partial)
        if not isinstance(out, list):
            out = [out]
        return reduce_func(out)

    def match_partial_path(
        self,
        partial_path: PathType | list[PathType],
        *args: str | int,
    ) -> list[tuple[str | int, ...]]:
        """Returns a list of all paths in the given data that match the given
        partial path. For example, if partial_path is ``(0, "key1")``, this will
        return all paths in the data that match ``...[0]["key1"]``. If fed a
        list of partial paths, this will return all paths that match any of the
        partial paths. Do not confuse tuples with lists, as they are treated
        differently.

        The ":" string can be added to the partial path input to match all list
        indices that match it. This is useful for JSONs with a repeating
        structure.

        The match_partial_path algorithm searches for all paths in a dictionary
        that match the given partial path. The ``partial_path`` can be a string,
        integer, a special ":" string, or a tuple of strings/integers that
        represents the  path to an item in the dictionary. For example, the path
        ``("key1", "key2", 2)`` corresponds to ``...["key1"]["key2"][2]`` in the
        dictionary. The algorithm returns a list of all paths that match the
        ``partial_path``. Each path in the result is represented as a tuple of
        strings and integers where strings correspond to dictionary keys and
        integers correspond to list indices.

        For instance, if ``data`` is:

        >>> data = QueryResponse({
        ...     "key1": {
        ...         "key2": [
        ...             {"key3": 1},
        ...             {"key3": 2},
        ...         ]
        ...     },
        ...     "key4": {
        ...         "key5": [
        ...             {"key3": 3},
        ...             {"key3": 4},
        ...         ]
        ...     }
        ... })

        Then ``data.match_partial_path((0, "key3"))`` will return:

        >>> [
        ...     ("key1", "key2", 0, "key3"),
        ...     ("key4", "key5", 0, "key3"),
        ... ]

        If ``data.match_partial_path("key3")`` is called, the result will be:

        >>> [
        ...     ("key1", "key2", 0, "key3"),
        ...     ("key1", "key2", 1, "key3"),
        ...     ("key4", "key5", 0, "key3"),
        ...     ("key4", "key5", 1, "key3"),
        ... ]

        If ``data.match_partial_path([(0, "key3"), "key2"])`` is called, the
        result will be:

        >>> [
        ...     ("key1", "key2", 0, "key3"),
        ...     ("key4", "key5", 0, "key3"),
        ...     ("key1", "key2"),
        ... ]

        If ``data.match_partial_path(("key1", "key2", ":", "key3"))`` is called,
        the result will be:

        >>> [
        ...     ("key1", "key2", 0, "key3"),
        ...     ("key1", "key2", 1, "key3"),
        ... ]

        Args:
            partial_path (PathType | list[PathType]): The partial path
                to match. If the partial path is a tuple, this method will treat
                it as a path. For example, a ``partial_path`` argument of ``(0,
                "key1")`` will be treated as ``...[0]["key1"]``. If the partial
                path is a string, this method will treat it as a key in a
                dictionary. Integers will be treated as indices in a list.
            *args (str | int): If *args is not empty, treat each arg as part of
                the partial path. For example, if
                ``data.match_partial_path(0, "key1")`` is called, the result
                is equivalent to ``data.match_partial_path((0, "key1"))``. Note
                that if *args is not empty, the ``partial_path`` argument must
                be a string or integer, not a tuple. Use the ``partial_path``
                argument with a list of paths instead of passing a PathType in
                *args.

        Raises:
            TypeError: If *args is not empty and the ``partial_path`` argument
                is a tuple. Use a list of paths instead of passing a PathType in
                *args.

        Returns:
            list[tuple[str | int, ...]]: A list of all paths that match the
                given partial path. Each path is represented as a tuple of
                strings and integers, where strings correspond  to dictionary
                keys and integers correspond to list indices.
        """
        # If *args is not empty, treat each arg as part of the partial path
        if args and isinstance(partial_path, tuple):
            raise TypeError(
                "Cannot use *args with a full path. Use a list of paths "
                "instead. *args is a shortcut syntax to not have to use a tuple"
                " for a single path."
            )
        elif args and not isinstance(partial_path, (tuple, list)):
            partial_path = cast(PathType, (partial_path, *args))

        return match_partial_path(self._data, partial_path)

    def get_partial_path(
        self,
        partial_path: PathType | list[PathType],
        *args: str | int,
    ) -> list[Any]:
        """Returns a list of values for all paths in the given data that match
        the provided partial path. This function first calls
        `match_partial_path` to find all matching paths and then retrieves the
        value at each of those paths using the `get` method.

        The input `partial_path` can be a single partial path or a list of
        partial paths, consisting of strings, integers, or a special colon (":")
        symbol, which represents matching all list indices.

        Example usage:

        >>> data = QueryResponse({
        ...     "key1": {
        ...         "key2": [
        ...             {"key3": 1},
        ...             {"key3": 2},
        ...         ]
        ...     },
        ...     "key4": {
        ...         "key5": [
        ...             {"key3": 3},
        ...             {"key3": 4},
        ...         ]
        ...     }
        ... })

        >>> data.get_partial_path("key3")
        [1, 2, 3, 4]

        Args:
            partial_path (PathType | list[PathType]): The partial path
                to match. If the partial path is a tuple, this method will treat
                it as a path. For example, a ``partial_path`` argument of ``(0,
                "key1")`` will be treated as ``...[0]["key1"]``. If the partial
                path is a string, this method will treat it as a key in a
                dictionary. Integers will be treated as indices in a list.
            *args (str | int): If *args is not empty, treat each arg as part of
                the partial path. For example, if
                ``data.get_partial_path(0, "key1")`` is called, the result
                is equivalent to ``data.get_partial_path((0, "key1"))``. Note
                that if *args is not empty, the ``partial_path`` argument must
                be a string or integer, not a tuple. Use the ``partial_path``
                argument with a list of paths instead of passing a PathType in
                *args.

        Returns:
            list[Any]: A list of values at all paths that match the
                given partial path.
        """
        paths = self.match_partial_path(partial_path, *args)
        return [self.get(path) for path in paths]

    def get(self, key: PathType, default: Any = None) -> Any:
        """Returns the value at the given key. If the key is not found, returns
        the default value.

        This method is similar to ``self[key]`` except that it will not raise
        an error if the key is not found. Instead, it will return the default
        value. This method will still raise a type error if a string key is
        given and the key's level data is a list.

        Args:
            key (PathType): The key to get the value at. If the key is a tuple,
                this method will treat it as taking multiple keys in order to
                get to the data.
            default (Any): The default value to return if the key is not found.
                Defaults to None.

        Returns:
            Any: The value at the given key or the default value if the key is
                not found.
        """
        try:
            return self[key]
        except (KeyError, IndexError):
            return default



import ast
import csv
import gzip
import hashlib
import json
from typing import Any, Callable, Literal, overload

from splatnet3_scraper.utils import delinearize_json, linearize_json


class LinearJSON:
    """Class containing methods for linearized JSON objects."""

    def __init__(
        self, header: list[str] | tuple[str, ...], data: list[list] | list
    ) -> None:
        """Initializes a LinearJSON.

        Args:
            header (list[str]): The header of the JSON object.
            data (list[Any]): The data of the JSON object.
        """
        self.header = header
        self.data = data if isinstance(data[0], list) else [data]
        self.__validate()

    @staticmethod
    def from_json(object: dict[str, Any]) -> "LinearJSON":
        """Creates a LinearJSON object from a JSON object.

        Args:
            object (dict[str, Any]): The JSON object to convert.

        Returns:
            LinearJSON: The LinearJSON object.
        """
        header, data = linearize_json(object)
        return LinearJSON(header, [data])

    @staticmethod
    def hash(obj: list[str] | tuple[str, ...]) -> str:
        """Returns the hash of the header.

        Args:
            obj (list[str] | tuple[str, ...]): The header to hash.

        Returns:
            str: The hash of the header.
        """
        header_tuple = tuple(obj)
        # Use a stable hash function to ensure the same hash is generated
        # across different Python versions
        return hashlib.sha256(str(header_tuple).encode()).hexdigest()

    def hashed_header(self) -> str:
        """Returns the hash of the header.

        Returns:
            str: The hash of the header.
        """
        return self.hash(self.header)

    def delinearize(self) -> dict[str, list[dict[str, Any]]]:
        """Delinearizes the JSON object into a big JSON object.

        Returns:
            dict[str, list[dict[str, Any]]]: The big JSON object.
        """
        out: dict[str, list[dict[str, Any]]] = {"data": []}
        for row in self.data:
            out["data"].append(delinearize_json(self.header, row))
        return out

    def __eq__(self, other: object) -> bool:
        """Returns whether the other object is equal to this object.

        Args:
            other (object): The other object to compare.

        Returns:
            bool: Whether the other object is equal to this object.
        """
        if isinstance(other, LinearJSON):
            return self.header == other.header and self.data == other.data
        elif isinstance(other, list):
            return self.header == other[0] and self.data == other[1:]
        return False

    def __validate(self) -> None:
        """Validates the header and data of the LinearJSON object.

        Raises:
            ValueError: If the header and data are not the same length.
        """
        for row in self.data:
            try:
                assert len(row) == len(self.header)
            except AssertionError:
                raise ValueError("The header and data are not the same length.")

    def __standardize_new_header(self, new_header: list[str]) -> None:
        """If the new header is different from the current header, this method
        will remove columns from the data that are not in the new header and add
        None values for columns that are in the new header but not the current
        header.

        Raises:
            ValueError: If the new header has duplicate columns.

        Args:
            new_header (list[str]): The new header.
        """
        new_header_columns = [x for x in new_header if x not in self.header]
        removed_header_columns = [x for x in self.header if x not in new_header]

        # Check if the new header has any duplicates and raise an error if so
        if len(new_header) != len(set(new_header)):
            raise ValueError("The new header has duplicate columns.")

        if (not new_header_columns) and (not removed_header_columns):
            return
        new_data = []
        for row in self.data:
            new_row: list[Any] = []
            for col in new_header:
                if col in new_header_columns:
                    new_row.append(None)
                elif col in removed_header_columns:
                    # Shouldn't happen, but making the logic explicit
                    continue
                else:
                    original_index = self.header.index(col)
                    new_row.append(row[original_index])

            new_data.append(new_row)
        self.header = new_header
        self.data = new_data

    @staticmethod
    def merge_headers(left: "LinearJSON", right: "LinearJSON") -> list[str]:
        """Merges the headers of the two given LinearJSON objects.

        Args:
            left (LinearJSON): The left LinearJSON object to merge.
            right (LinearJSON): The right LinearJSON object to merge.

        Returns:
            list[str]: The new header.
        """
        new_header_set = set(left.header).union(set(right.header))
        new_header = list(new_header_set)
        new_header = sorted(new_header, key=lambda x: (len(x), x))
        return new_header

    def append(self, other: "LinearJSON") -> None:
        """Appends a LinearJSON object to this one. If the headers are not the
        same, the headers will be merged and the data will be standardized to
        the new header.

        Args:
            other (LinearJSON): The other LinearJSON object to append.
        """
        if self.header != other.header:
            new_header = self.merge_headers(self, other)
            # New header will have new columns anywhere in the header, so we
            # need to identify the new columns and insert None values for them
            # in the data
            obj_list: list[LinearJSON] = [self, other]
            for obj in obj_list:
                obj.__standardize_new_header(new_header)
        self.data.extend(other.data)

    def transpose(self) -> list[list[Any]]:
        """Transposes the data.

        Returns:
            list[list[Any]]: The transposed data.
        """
        return [list(x) for x in zip(*self.data)]

    @overload
    def stringify(self, include_header: Literal[True] = ...) -> tuple[str, str]:
        ...

    @overload
    def stringify(self, include_header: Literal[False]) -> str:
        ...

    @overload
    def stringify(self, include_header: bool) -> str | tuple[str, str]:
        ...

    def stringify(
        self,
        include_header: bool = True,
    ) -> str | tuple[str, str]:
        """Stringifies the LinearJSON object. If include_header is True, then
        the header and data are returned as a tuple. Otherwise, only the data
        is returned as a string.

        Args:
            include_header (bool): Whether to include the header. Defaults to
                True.

        Returns:
            str | tuple[str, str]: The header and data as a string, or just the
                data as a string.
        """
        data = self.data
        data_str = []
        for row in data:
            out_row = []
            # If the row has a comma, then we need to wrap the row in quotes
            for col in row:
                if isinstance(col, str) and "," in col:
                    out_col = f'"{col}"'
                else:
                    out_col = str(col)
                out_row.append(out_col)
            data_str.append(",".join(out_row))
        data_str_out = "\n".join(data_str)
        if include_header:
            header_str = ",".join(self.header)
            return header_str, data_str_out
        return data_str_out

    def remove_columns(self, columns: list[str]) -> None:
        """Removes columns from the LinearJSON object.

        Args:
            columns (list[str]): The columns to remove. If a column is not in
                the header, it will be ignored.
        """
        new_header = [x for x in self.header if x not in columns]
        self.__standardize_new_header(new_header)

    def remove_url_columns(self) -> None:
        """Removes columns that are URLs from the LinearJSON object."""
        url_columns = [x for x in self.header if x.lower().endswith("url")]
        self.remove_columns(url_columns)


class JSONParser:
    """Class containing JSON methods for saving and loading data."""

    def __init__(self, data: dict[str, Any] | list[dict[str, Any]]) -> None:
        """Initializes a JSONParser.

        Args:
            data (dict[str, Any] | list[dict[str, Any]]): The data to parse.
        """
        if isinstance(data, dict):
            data = [data]
        self.data = data

    def __len__(self) -> int:
        return len(self.data)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, JSONParser):
            return self.data == other.data
        return False

    def __repr__(self) -> str:
        return f"JSONParser({len(self)} rows)"

    def __to_linear_json(self) -> LinearJSON:
        """Converts the JSON object to a LinearJSON object.

        Returns:
            LinearJSON: The LinearJSON object.
        """
        header, data = linearize_json(self.data[0])
        out = LinearJSON(header, data)
        for row in self.data[1:]:
            header, data = linearize_json(row)
            out.append(LinearJSON(header, data))
        return out

    def remove_columns(self, columns: list[str]) -> None:
        """Removes columns from the data.

        Args:
            columns (list[str]): The columns to remove.
        """
        linear_json = self.__to_linear_json()
        linear_json.remove_columns(columns)
        self.data = linear_json.delinearize()["data"]

    def remove_url_columns(self) -> None:
        """Removes URL columns from the data."""
        linear_json = self.__to_linear_json()
        linear_json.remove_url_columns()
        self.data = linear_json.delinearize()["data"]

    def to_csv(self, path: str) -> None:
        """Saves the JSON object to a CSV file.

        Given the JSON Parser's data, this method will save the data to a CSV by
        linearizing the JSON object. Given the following example JSON object:

        >>> json = {
        ...     "a": 1,
        ...     "b": {
        ...         "c": 2,
        ...         "d": 3,
        ...     },
        ...     "e": [4, 5, 6],
        ...     "f": [
        ...         {"g": 7, "h": 8},
        ...         {"g": 9, "h": 10},
        ...     ],
        ... }

        The following CSV file will be created with spaces added for clarity:

        >>> "a, b.c, b.d, e;0, e;1, e;2, f;0.g, f;0.h, f;1.g, f;1.h"
        ... "1, 2, 3, 4, 5, 6, 7, 8, 9, 10"

        Args:
            path (str): The path to save the CSV file to.
        """
        linear_json = self.__to_linear_json()
        with open(path, "w", encoding="utf-8") as f:
            header, data = linear_json.stringify()
            f.write(header + "\n")
            f.write(data)

    def __to_json(self, path: str, use_gzip: bool, **kwargs) -> None:
        """Saves the JSON object to a JSON file. Any keyword arguments are
        passed to the json.dump method. If gzip is True, the file will be
        compressed with gzip.

        Args:
            path (str): The path to save the JSON file to.
            use_gzip (bool): Whether or not to compress the file with gzip.
            **kwargs: Any keyword arguments to pass to the json.dump method.
        """
        default_kwargs: dict[str, Any] = {"indent": 4}
        default_kwargs.update(kwargs)
        if use_gzip:
            open_function: Callable = gzip.open
            open_kwargs = {"mode": "wt", "encoding": "utf-8"}
        else:
            open_function = open
            open_kwargs = {"mode": "w", "encoding": "utf-8"}

        with open_function(path, **open_kwargs) as f:
            json.dump(self.data, f, **default_kwargs)

    def to_json(self, path: str, **kwargs) -> None:
        """Saves the JSON object to a JSON file. Any keyword arguments are
        passed to the json.dump method.

        Args:
            path (str): The path to save the JSON file to.
            **kwargs: Keyword arguments to pass to the json.dump method.
        """
        self.__to_json(path, False, **kwargs)

    def to_gzipped_json(self, path: str, **kwargs) -> None:
        """Saves the JSON object to a gzipped JSON file. Any keyword arguments
        are passed to the json.dump method.

        Args:
            path (str): The path to save the gzipped JSON file to.
            **kwargs: Keyword arguments to pass to the json.dump method.
        """
        self.__to_json(path, True, **kwargs)

    def to_parquet(self, path: str, **kwargs) -> None:
        """Saves the JSON object to a Parquet file. Any keyword arguments are
        passed to the pandas.DataFrame.to_parquet method.

        Args:
            path (str): The path to save the Parquet file to.
            **kwargs: Keyword arguments to pass to the pq.write_table method.

        Raises:
            ImportError: If the parquet extra is not installed.
        """
        try:
            import numpy as np
            import pyarrow as pa
            import pyarrow.parquet as pq
        except ImportError:
            raise ImportError(
                "parquet format requires the [parquet] extra to be installed. "
                'Try "pip install splatnet3_scraper[parquet]" or "poetry '
                'install --extras parquet" if you are developing.'
            )
        linear_json = self.__to_linear_json()
        numpy_data = np.array(linear_json.data)
        arrays = [pa.array(data) for data in numpy_data.T]
        del numpy_data
        table = pa.Table.from_arrays(arrays, names=linear_json.header)
        pq.write_table(table, path, **kwargs)

    @staticmethod
    def automatic_type_conversion(row: list[str]) -> list[Any]:
        """Converts a row of strings to the most appropriate type.

        Args:
            row (list[str]): The row of strings to convert.

        Returns:
            list[Any]: The converted row.
        """
        converted_row = []
        for col in row:
            try:
                value = ast.literal_eval(col)
            except (ValueError, SyntaxError):
                value = None if col == "" else col
            converted_row.append(value)
        return converted_row

    @classmethod
    def from_csv(cls, path: str) -> "JSONParser":
        """Loads a JSON object from a CSV file.

        Args:
            path (str): The path to load the CSV file from.

        Returns:
            JSONParser: The JSONParser object.
        """
        with open(path, "r", encoding="utf-8") as f:
            reader = csv.reader(
                f,
                quotechar='"',
                delimiter=",",
                quoting=csv.QUOTE_MINIMAL,
                skipinitialspace=True,
            )
            header = next(reader)
            data = [JSONParser.automatic_type_conversion(row) for row in reader]
        delinearized_data = [delinearize_json(header, row) for row in data]
        return cls(delinearized_data)

    @classmethod
    def from_json(cls, path: str) -> "JSONParser":
        """Loads a JSON object from a JSON file.

        Args:
            path (str): The path to load the JSON file from.

        Returns:
            JSONParser: The JSONParser object.
        """
        with open(path, "r", encoding="utf-8") as f:
            return cls(json.load(f))

    @classmethod
    def from_gzipped_json(cls, path: str) -> "JSONParser":
        """Loads a JSON object from a gzipped JSON file.

        Args:
            path (str): The path to load the gzipped JSON file from.

        Returns:
            JSONParser: The JSONParser object.
        """
        with gzip.open(path, "rt", encoding="utf-8") as f:
            return cls(json.load(f))

    @classmethod
    def from_parquet(cls, path: str) -> "JSONParser":
        """Loads a JSON object from a Parquet file.

        Args:
            path (str): The path to load the Parquet file from.

        Raises:
            ImportError: If the parquet extra is not installed.

        Returns:
            JSONParser: The JSONParser object.
        """
        try:
            import numpy as np
            import pyarrow.parquet as pq
        except ImportError:
            raise ImportError(
                "parquet format requires the [parquet] extra to be installed. "
                'Try "pip install splatnet3_scraper[parquet]" or "poetry '
                'install --extras parquet" if you are developing.'
            )
        table = pq.read_table(path)
        data = [table.column(i).to_numpy() for i in range(len(table.columns))]
        out_data = np.array(data).T
        header = [field.name for field in table.schema]
        return cls(delinearize_json(header, out_data.tolist()))



from __future__ import annotations

import json
from typing import Any

import requests

from splatnet3_scraper.constants import (
    DEFAULT_USER_AGENT,
    GRAPHQL_URL,
    SPLATNET_URL,
)
from splatnet3_scraper.utils import get_splatnet_hashes, get_splatnet_version


class GraphQLQueries:
    """The GraphQLQueries class that contains the GraphQL queries used by
    ``splatnet3_scraper``. The Nintendo Switch Online API only allows persistent
    queries, so making queries requires using the hashes for said queries. The
    great folks over at `imink` have a GitHub Actions workflow that checks for
    and updates the hashes for the queries every five minutes. The hashes are
    stored in a JSON file that is hosted on GitHub. This class automatically
    downloads the JSON file and stores the hashes in a dictionary. The hashes
    are then used to make queries to the Nintendo Switch Online API and can be
    accessed using their string names.
    """

    def __init__(self) -> None:
        """Initializes the GraphQLQueries class. Initializes a requests.Session
        and stores it in the session attribute. Also gets the hashes for the
        GraphQL queries and stores them in the hash_map attribute. The hashes
        are stored in a dictionary where the keys are the names of the queries
        and the values are the hashes.
        """
        self.session = requests.Session()

    def get_query(self, query_name: str) -> str:
        """Gets a GraphQL query hash given the name of the query.

        Uses the hash_map attribute to get the hash for the query, which is
        then returned. For more information on the valid queries, see the
        ``queries`` page of the documentation.

        Args:
            query_name (str): The name of the query.

        Returns:
            str: The GraphQL query hash.
        """
        return get_splatnet_hashes()[query_name]

    def query_header(
        self,
        bullet_token: str,
        language: str,
        user_agent: str | None = None,
        override: dict[str, str] = {},
    ) -> dict[str, str]:
        """Generates the headers that are used for the GraphQL queries made for
        the SplatNet 3 API.

        The headers are generated using the bullet token, language and user
        agent. Any headers that are passed in the override parameter will
        override the default headers. An example of the headers that are
        generated is shown below.

        >>> headers = {
        ...     "Authorization": f"Bearer {bullet_token}",
        ...     "Accept-Language": language,
        ...     "User-Agent": user_agent,
        ...     "X-Web-View-Ver": web_version,
        ...     "Content-Type": "application/json",
        ...     "Accept": "*/*",
        ...     "Origin": SPLATNET_URL,
        ...     "X-Requested-With": "com.nintendo.znca",
        ...     "Referer": (
        ...         f"{SPLATNET_URL}?"
        ...         f"lang={language}"
        ...         f"&na_country={language[-2:]}"
        ...         f"&na_lang={language}"
        ...     ),
        ...     "Accept-Encoding": "gzip, deflate",
        ... }

        Args:
            bullet_token (str): The bullet token.
            language (str): The language code to use, for example, "en-US".
            user_agent (str | None): The user agent to use. If None, the default
                user agent will be used. Defaults to None.
            override (dict[str, str]): Any headers that should override the
                default headers. Defaults to {}.

        Returns:
            dict[str, str]: The headers for the GraphQL queries.
        """

        if user_agent is None:
            user_agent = DEFAULT_USER_AGENT

        headers = {
            "Authorization": f"Bearer {bullet_token}",
            "Accept-Language": language,
            "User-Agent": user_agent,
            "X-Web-View-Ver": get_splatnet_version(),
            "Content-Type": "application/json",
            "Accept": "*/*",
            "Origin": SPLATNET_URL,
            "X-Requested-With": "com.nintendo.znca",
            "Referer": (
                f"{SPLATNET_URL}?"
                f"lang={language}"
                f"&na_country={language[-2:]}"
                f"&na_lang={language}"
            ),
            "Accept-Encoding": "gzip, deflate",
        }
        headers.update(override)
        return headers

    def query_body_hash(
        self, query_hash: str | bytes, variables: dict[str, Any] = {}
    ) -> str:
        """Generates the body for the GraphQL queries, as a string.

        The body is generated using the query hash and the variables that are
        passed in. The body is a JSON string that contains the query hash and
        the variables. An example of the body that is generated is shown
        below.

        >>> body = {
        ...     "extensions": {
        ...         "persistedQuery": {
        ...             "sha256Hash": query_hash,
        ...             "version": 1,
        ...         }
        ...     },
        ...     "variables": variables,
        ... }

        Args:
            query_hash (str | bytes): The hash of the query.
            variables (dict[str, Any]): The variables to pass to the query. If
                the query does not take any variables, this can be an empty
                dictionary. Defaults to {}.

        Returns:
            str: The body for the GraphQL queries, as a string.
        """
        out = {
            "extensions": {
                "persistedQuery": {
                    "sha256Hash": query_hash,
                    "version": 1,
                }
            },
            "variables": variables,
        }
        return json.dumps(out)

    def query_body(
        self, query_name: str, variables: dict[str, str] = {}
    ) -> str:
        """Generates the body for the GraphQL queries, as a string.

        This method is a wrapper around the ``query_body_hash`` method. It gets
        the query hash using the ``get_query`` method and then calls the
        ``query_body_hash`` method to generate the body. For more information on
        the valid queries, see the `queries` page of the documentation.

        Args:
            query_name (str): The name of the query.
            variables (dict[str, str]): The variables to pass to the query. If
                the query does not take any variables, this can be an empty
                dictionary. Defaults to {}.

        Returns:
            str: The body for the GraphQL queries, as a string.
        """
        query = self.get_query(query_name)
        return self.query_body_hash(query, variables)

    def query_hash(
        self,
        query_hash: str | bytes,
        bullet_token: str,
        gtoken: str,
        language: str,
        user_agent: str | None = None,
        variables: dict[str, Any] = {},
        override: dict[str, Any] = {},
    ) -> requests.Response:
        """Makes a GraphQL query using the persisted query hash. This method
        generates the headers and body for the query and then makes the
        request. For more information on the valid queries, see the `queries`
        page of the documentation.

        Args:
            query_hash (str | bytes): The hash of the query.
            bullet_token (str): The bullet token.
            gtoken (str): The gtoken.
            language (str): The language code to use.
            user_agent (str | None): The user agent to use. If None, the default
                user agent will be used. Defaults to None.
            variables (dict[str, Any]): The variables to pass to the query. If
                the query does not take any variables, this can be an empty
                dictionary. Defaults to an empty dictionary.
            override (dict[str, Any]): Any headers that should override the
                default headers. Defaults to an empty dictionary.

        Returns:
            requests.Response: The response from the GraphQL query.
        """
        header = self.query_header(bullet_token, language, user_agent, override)
        body = self.query_body_hash(query_hash, variables)
        cookies = {
            "_gtoken": gtoken,
        }
        return self.session.post(
            GRAPHQL_URL, headers=header, data=body, cookies=cookies
        )

    def query(
        self,
        query_name: str,
        bullet_token: str,
        gtoken: str,
        language: str,
        user_agent: str | None = None,
        variables: dict[str, Any] = {},
        override: dict[str, Any] = {},
    ) -> requests.Response:
        """Makes a GraphQL query. This method is a wrapper around the
        ``query_hash`` method. It gets the query hash using the ``get_query``
        method and then calls the ``query_hash`` method to make the request. For
        more information on the valid queries, see the `queries` page of the
        documentation.

        Args:
            query_name (str): The name of the query.
            bullet_token (str): The bullet token.
            gtoken (str): The gtoken.
            language (str): The language code to use.
            user_agent (str | None): The user agent to use. If None, the default
                user agent will be used. Defaults to None.
            variables (dict[str, Any]): The variables to pass to the query. If
                the query does not take any variables, this can be an empty
                dictionary. Defaults to an empty dictionary.
            override (dict[str, Any]): Any headers that should override the
                default headers. Defaults to an empty dictionary.

        Returns:
            requests.Response: The response from the GraphQL query.
        """
        query = self.get_query(query_name)
        return self.query_hash(
            query,
            bullet_token,
            gtoken,
            language,
            user_agent,
            variables,
            override,
        )


queries = GraphQLQueries()



import configparser
from typing import Literal, overload

from splatnet3_scraper.auth import (
    EnvironmentVariablesManager,
    Token,
    TokenManager,
)
from splatnet3_scraper.constants import DEFAULT_USER_AGENT, IMINK_URL


class Config:
    """Class that can access the token manager as well as additional options."""

    def __init__(
        self,
        config_path: str | None = None,
        *args,
        token_manager: TokenManager | None = None,
    ) -> None:
        """Initializes the class. If token_manager is given, it will assume that
        this is a first time initialization and has not been setup yet.

        Token manager will look for tokens in the following order:
            1. the config_path argument
            2. check the current working directory for ".splatnet3_scraper"
            3. check for environment variables for defined tokens
            4. check the current working directory for "tokens.ini"

        If none of these are found, an exception will be raised.

        Args:
            config_path (str | None): The path to the config file. If None, it
                will look for ".splatnet3_scraper" in the current working
                directory. Defaults to None.
            *args: These are ignored.
            token_manager (TokenManager | None): The token manager to use.
                Keyword argument. If given, it will skip the post-init method.
                Defaults to None.
        """
        self.additional_accepted_options: list[str] = []
        self.additional_deprecated_options: dict[str, str] = {}
        self.additional_default_options: dict[str, str] = {}

        if token_manager is None:
            self.generate_token_manager(config_path)
            return
        else:
            self.config_path = config_path

        self.token_manager = token_manager
        self.initialize_options()

    def generate_token_manager(self, config_path: str | None = None) -> None:
        """Generates the token manager.

        Will only be called if the token manager is not given in the constructor
        or if the ``token_manager`` argument in the constructor is None. This
        means that the user prefers to use the default loading method for the
        token manager, which is to look for tokens in the following order:

            1. the config_path argument
            2. check the current working directory for ".splatnet3_scraper"
            3. check for environment variables for defined tokens
            4. check the current working directory for "tokens.ini"

        If none of these are found, an exception will be raised.

        Args:
            config_path (str | None): The path to the config file. If None, it
                will look for ".splatnet3_scraper" in the current working
                directory.
        """
        if config_path is not None:
            self.token_manager = TokenManager.from_config_file(config_path)
        # cgarza: A little bit of redundancy here, need better method.
        else:
            self.token_manager = TokenManager.load()

        config_path = (
            ".splatnet3_scraper" if config_path is None else config_path
        )

        self.config_path = config_path
        self.config = configparser.ConfigParser()
        self.config.read(config_path)
        try:
            self.options = self.config.options("options")
        except configparser.NoSectionError:
            self.config.add_section("options")
            self.options = self.config.options("options")
        self.manage_options()

        with open(config_path, "w") as configfile:
            self.config.write(configfile)

    def initialize_options(self) -> None:
        if getattr(self, "config", None) is not None:
            return

        origin = self.token_manager.origin
        self.config = configparser.ConfigParser()
        self.config.add_section("options")
        self.options = self.config.options("options")

        if origin["origin"] == "env":
            # If the tokens are in environment variables, we should check for
            # any additional tokens in the token's environment manager and add
            # them to the config variable under the "options" section.
            tokens = self.token_manager.env_manager.get_all()
            for token, value in tokens.items():
                if token in self.token_manager.env_manager.BASE_TOKENS:
                    continue
                if value is None:
                    continue
                self.config["options"][token] = value

            return

    @classmethod
    def from_env(
        cls, env_manager: EnvironmentVariablesManager | None = None
    ) -> "Config":
        """Creates a Config instance using the environment variables.

        Args:
            env_manager (EnvironmentVariablesManager | None): The environment
                variables manager to use. If None, it will create a new one.
                Defaults to None.

        Returns:
            Config: The Config instance.
        """
        return cls(token_manager=TokenManager.from_env(env_manager))

    @classmethod
    def from_s3s_config(cls, config_path: str) -> "Config":
        """Creates a Config instance using the config file from s3s.

        Args:
            config_path (str): The path to the config file.

        Returns:
            Config: The Config instance.
        """
        return cls(token_manager=TokenManager.from_text_file(config_path))

    def save(
        self, path: str | None = None, include_tokens: bool = True
    ) -> None:
        """Saves the config file to the given path.

        Args:
            path (str | None): The path to save the config file to. If the token
                manager is using environment variables, the tokens section will
                be removed from the config file. If None, it will save to the
                path given in the constructor or ".splatnet3_scraper" in the
                current working directory.
            include_tokens (bool): Whether or not to include the tokens in the
                config file. If False, the tokens will be removed from the
                config file.
        """
        # Check if the user has the tokens in a separate file
        origin = self.token_manager._origin["origin"]
        if (origin == "env") or (not include_tokens):
            # Remove the token manager from the config file
            self.config.remove_section("tokens")
        if path is None and self.config_path is not None:
            path = self.config_path
        elif path is None:
            path = ".splatnet3_scraper"

        with open(path, "w") as configfile:
            self.config.write(configfile)

    def manage_options(self) -> None:
        """Manage the options in the config file.

        This function will move invalid options to the "unknown" section and
        move deprecated options to the "deprecated" section while replacing them
        with the new option name.
        """
        for option in self.options:
            if option not in (
                self.ACCEPTED_OPTIONS + list(self.DEPRECATED_OPTIONS.keys())
            ):
                if not self.config.has_section("unknown"):
                    self.config.add_section("unknown")
                self.config["unknown"][option] = self.config["options"][option]
                self.config.remove_option("options", option)
            if option in self.DEPRECATED_OPTIONS:
                deprecated_name = option
                option_name = self.DEPRECATED_OPTIONS[option]
                if not self.config.has_section("deprecated"):
                    self.config.add_section("deprecated")
                # Make a copy of the deprecated option in the deprecated section
                # and then replace the deprecated option with the new option
                self.config["deprecated"][deprecated_name] = self.config[
                    "options"
                ][deprecated_name]
                self.config["options"][option_name] = self.config["options"][
                    deprecated_name
                ]
                self.config.remove_option("options", option)

    _ACCEPTED_OPTIONS = [
        "user_agent",
        "log_level",
        "log_file",
        "export_path",
        "language",
        "lang",
        "country",
        "stat.ink_api_token",
        "stat_ink_api_token",
        "statink_api_token",
        "f_token_url",
    ]

    _DEPRECATED_OPTIONS = {
        "api_key": "stat.ink_api_token",
        "f_gen": "f_token_url",
    }

    _DEFAULT_OPTIONS = {
        "user_agent": DEFAULT_USER_AGENT,
        "f_gen": IMINK_URL,
        "export_path": "",
    }

    @property
    def ACCEPTED_OPTIONS(self) -> list[str]:
        return self._ACCEPTED_OPTIONS + self.additional_accepted_options

    @property
    def DEPRECATED_OPTIONS(self) -> dict[str, str]:
        return {
            **self._DEPRECATED_OPTIONS,
            **self.additional_deprecated_options,
        }

    @property
    def DEFAULT_OPTIONS(self) -> dict[str, str]:
        return {**self._DEFAULT_OPTIONS, **self.additional_default_options}

    def get(self, key: str) -> str:
        """Get the value of an option. If the option is not set, the default
        value will be returned.

        Args:
            key (str): The name of the option.

        Raises:
            KeyError: If the option is valid, but not set and has no default.
            KeyError: If the option is not valid.

        Returns:
            str: The value of the option.
        """
        if key in self.ACCEPTED_OPTIONS:
            if key in self.config["options"]:
                return self.config["options"][key]
            elif key in self.DEFAULT_OPTIONS:
                return self.DEFAULT_OPTIONS[key]
            else:
                raise KeyError(f"Option not set and has no default: {key}")
        elif key in self.DEPRECATED_OPTIONS:
            return self.get(self.DEPRECATED_OPTIONS[key])
        else:
            raise KeyError(f"Invalid option: {key}")

    def get_data(self, key: str) -> str:
        if not self.config.has_section("data"):
            self.config.add_section("data")
            data = self.token_manager.data
            for k, v in data.items():
                self.config["data"][k] = v
        return self.config["data"][key]

    @overload
    def get_token(self, key: str, full_token: Literal[False] = ...) -> str:
        ...

    @overload
    def get_token(self, key: str, full_token: Literal[True]) -> Token:
        ...

    @overload
    def get_token(self, key: str, full_token: bool) -> str | Token:
        ...

    def get_token(self, key: str, full_token: bool = False) -> str | Token:
        """Get the value of a token.

        Args:
            key (str): The name of the token.
            full_token (bool): Whether or not to return the full token. If
                False, only the token value will be returned.

        Returns:
            str: The value of the token.
        """
        return self.token_manager.get(key, full_token)

    def add_accepted_options(self, options: list[str]) -> None:
        """Add options to the list of accepted options.

        Args:
            options (list[str]): The list of options to add.
        """
        self.additional_accepted_options.extend(options)

    def add_deprecated_options(self, options: dict[str, str]) -> None:
        """Add options to the list of deprecated options.

        Args:
            options (dict[str, str]): The list of options to add.
        """
        self.additional_deprecated_options.update(options)

    def add_default_options(self, options: dict[str, str]) -> None:
        """Add options to the list of default options.

        Args:
            options (dict[str, str]): The list of options to add.
        """
        self.additional_default_options.update(options)

    def remove_accepted_options(self, options: list[str]) -> None:
        """Remove options from the list of accepted options.

        Args:
            options (list[str]): The list of options to remove.
        """
        for option in options:
            if option in self.additional_accepted_options:
                self.additional_accepted_options.remove(option)
            else:
                self._ACCEPTED_OPTIONS.remove(option)

    def remove_deprecated_options(self, options: list[str]) -> None:
        """Remove options from the list of deprecated options.

        Args:
            options (list[str]): The list of options to remove.
        """
        for option in options:
            if option in self.additional_deprecated_options:
                self.additional_deprecated_options.pop(option)
            else:
                self._DEPRECATED_OPTIONS.pop(option)

    def remove_default_options(self, options: list[str]) -> None:
        """Remove options from the list of default options.

        Args:
            options (list[str]): The list of options to remove.
        """
        for option in options:
            if option in self.additional_default_options:
                self.additional_default_options.pop(option)
            else:
                self._DEFAULT_OPTIONS.pop(option)



import json

import requests

from splatnet3_scraper.auth import NSO, TokenManager
from splatnet3_scraper.auth.exceptions import SplatNetException
from splatnet3_scraper.auth.graph_ql_queries import queries
from splatnet3_scraper.constants import TOKENS
from splatnet3_scraper.query.config import Config
from splatnet3_scraper.query.responses import QueryResponse
from splatnet3_scraper.utils import retry


class QueryHandler:
    """The QueryHandler class is the main class of the
    ``splatnet3_scraper.query`` module. It abstracts away the underlying
    implementation details of making queries to the SplatNet 3 API and makes it
    extremely easy to use. Token management is also handled by the QueryHandler
    so the user does not have to worry about having to regenerate tokens when
    they expire. The only exception to this is the session token, which the user
    must handle themselves as it is not possible to regenerate it automatically.
    The QueryHandler class is initialized with a Config class, which contains
    all the configuration options that the user can set. The QueryHandler class
    also contains multiple factory methods to create a new instance of the class
    to make it setting up the Config class much easier. The primary mode of
    operating the class is through the ``query`` method, which takes a query
    name and a dictionary of arguments and returns a QueryResponse object.
    """

    def __init__(self, config: Config) -> None:
        """Initializes the class, it is not meant to be instantiated directly,
        but rather through one of the available factory methods. Still, it can
        be instantiated directly if the user wants to use a custom Config class.

        Args:
            config (Config): The Config object to use. This object contains all
                the configuration options that the user can set, and the
                QueryHandler class will use these options as they are needed
                throughout the operation of the class. It is recommended that
                the user uses the factory methods to create a new instance of
                the class instead of instantiating it directly, as the factory
                methods will automatically create a Config object for the user
                based on the instantiation method they prefer and pass it to the
                constructor.
        """
        self.config = config

    @staticmethod
    def from_config_file(
        config_path: str | None = None,
    ) -> "QueryHandler":
        """Creates a new instance of the class using a configuration file.

        If the user does not provide a configuration file path, the default path
        will be used, which is ``.splatnet3_scraper`` in the user's current
        working directory. The configuration file this method accepts is one
        that aligns with the standard configuration file format used by the
        standard library ``configparser`` module. The configuration file must
        have a ``[Tokens]`` section, which only requires the ``session_token``
        option to be set. The ``session_token`` option must be set to the user's
        session token. For a full list of all options that can be set in the
        config file, see the documentation for the ``Config`` class.

        Args:
            config_path (str | None): The path to the configuration file. If
                None, the default configuration file path of
                ``.splatnet3_scraper`` in the user's current working directory
                will be used. Defaults to None.

        Returns:
            QueryHandler: A new instance of the class using the configuration
                file provided, with all the options set in the configuration
                file.
        """
        config = Config(config_path)
        return QueryHandler(config)

    @staticmethod
    def generate_session_token_url(
        user_agent: str | None = None,
    ) -> tuple[str, bytes, bytes]:
        """Generates the URL to use to get the session token.

        This method is used to generate the URL to use to get the session token
        from the Nintendo Switch Online website. The user must then visit this
        URL in their browser and log in to their Nintendo account. Once they
        have logged in, they will be redirected to a page that contains the user
        profile to select. The user must then select their user profile and
        then copy the URL from the "select this account" button. The URI of the
        button will contain the session token code, which the user must then
        pass to the ``parse_npf_uri`` method to get the session token. This
        method is currently incomplete and will be updated in the future with
        its counterpart method to parse the URI and return a ``QueryHandler``
        object with a valid session token already set in the ``Config`` object
        and saved to the default configuration path in the user's current
        working directory.

        Args:
            user_agent (str | None): The user agent to use when making the
                initial requests to the Nintendo Switch Online service. If None,
                the default user agent will be used. It is not recommended to
                change this unless you know what you are doing. Defaults to
                None.

        Returns:
            str: The URL to use to get the session token.
            bytes: The state ID.
            bytes: The challenge solution.
        """
        nso = NSO.new_instance()
        url = nso.generate_login_url(user_agent)
        return url, nso.state, nso.verifier

    @staticmethod
    def from_session_token(session_token: str) -> "QueryHandler":
        """Creates a new instance of the class using a session token.

        Given a session token, this method will create a new instance of the
        class with a valid session token already set in the ``Config`` object.
        This method is useful if the user already has a session token and does
        not want to generate an accompanying ``GTOKEN`` and ``BULLET_TOKEN``.

        Args:
            session_token (str): The session token to use. This token must be
                valid and not expired or revoked. If the token is invalid, the
                user will not be able to make any queries to the SplatNet 3 API.

        Returns:
            QueryHandler: A new instance of the class with a valid session
                token already set in the ``Config`` object. The ``GTOKEN`` and
                ``BULLET_TOKEN`` will also have been generated and set in the
                ``Config`` object.
        """
        token_manager = TokenManager.from_session_token(session_token)
        token_manager.generate_all_tokens()
        config = Config(token_manager=token_manager)
        return QueryHandler(config)

    @staticmethod
    def from_env() -> "QueryHandler":
        """Creates a new instance of the class using the environment
        variables.

        This method will create a new instance of the class using the
        environment variables. This method is useful if the user wants to start
        up the class in a CI/CD pipeline, a Docker container, or any other
        environment where the user does not want to store the session token in
        a file. The environment variables that this method will use are
        prepended by ``SN3S_`` and are all uppercase. The three environment
        variables supported are:

        - ``SN3S_SESSION_TOKEN``: The session token to use. This token must be
            valid and not expired or revoked.
        - ``SN3S_GTOKEN``: The GTOKEN to use.
        - ``SN3S_BULLET_TOKEN``: The BULLET_TOKEN to use.

        Returns:
            QueryHandler: A new instance of the class with a valid session
                token already set in the ``Config`` object. The ``Config``
                object is also flagged as using environment variables, so the
                user does not need to worry about accidentally saving the
                session token to disk.
        """
        config = Config.from_env()
        return QueryHandler(config)

    @staticmethod
    def from_s3s_config(path: str) -> "QueryHandler":
        """Creates a new instance of the class using the ``s3s`` config file.

        This method will create a new instance of the class by reading the
        ``s3s`` config file. The ``s3s`` config file is a file that is used by
        ``s3s`` to store the user's data. This method is useful if the user
        already has a ``s3s`` config file. The ``s3s`` config file is a JSON
        file that contains the user's session token, G token, bullet token, and
        other options and data.

        Args:
            path (str): The path to the s3s config file. This file must be a
                valid JSON file.

        Returns:
            QueryHandler: A new instance of the class from the ``s3s`` config
                file provided.
        """
        config = Config.from_s3s_config(path)
        return QueryHandler(config)

    def __query(
        self, query_name: str, variables: dict = {}
    ) -> requests.Response:
        """Internal method to query Splatnet 3, it does all of the heavy lifting
        and is used by the other methods to get the data.

        Args:
            query_name (str): The name of the query to use.
            variables (dict): The variables to use in the query. Defaults to {}.

        Returns:
            requests.Response: The response from the query.
        """
        return queries.query(
            query_name,
            self.config.get_token(TOKENS.BULLET_TOKEN),
            self.config.get_token(TOKENS.GTOKEN),
            self.config.get_data("language"),
            self.config.get("user_agent"),
            variables=variables,
        )

    # Repeat code, but I've elected to do this to make it easier to read
    def __query_hash(
        self, query_hash: str, variables: dict = {}
    ) -> requests.Response:
        """Internal method to query Splatnet 3, it does all of the heavy lifting
        and is used by the other methods to get the data.

        Args:
            query_hash (str): The hash of the query to use.
            variables (dict): The variables to use in the query. Defaults to {}.

        Returns:
            requests.Response: The response from the query.
        """
        return queries.query_hash(
            query_hash,
            self.config.get_token(TOKENS.BULLET_TOKEN),
            self.config.get_token(TOKENS.GTOKEN),
            self.config.get_data("language"),
            self.config.get("user_agent"),
            variables=variables,
        )

    # Repeat code, but I've elected to do this to make it easier to read
    @retry(times=1, exceptions=ConnectionError)
    def query_hash(
        self, query_hash: str, variables: dict = {}
    ) -> QueryResponse:
        """Given a query hash, it will query SplatNet 3 and return the response.

        A convenience function that will use a GraphQL query hash rather than a
        query name to query SplatNet 3. This method will automatically retry
        once if the query fails and will regenerate the tokens if the query
        fails for whatever reason. Some queries require variables to be passed
        in, and this method will allow the user to pass in those variables. It
        is not recommended to use this method unless the user knows what they
        are doing, as the query hashes are not stable and can and do change
        often and without warning. The user should use the ``query`` method
        instead as it will compensate for any changes to the query hashes.

        Args:
            query_hash (str): The query hash to use. This hash must be valid,
                and must be a string representation rather than a byte
                representation.
            variables (dict): The variables to use in the query. Some queries do
                not require variables and so this argument can be omitted. If
                the query does require variables and this argument is omitted,
                then the query will fail and a ``SplatNetException`` will be
                raised with the error message from SplatNet 3. Defaults to {}.

        Raises:
            SplatNetException: If the query is successful but returns a JSON
                object with an ``errors`` key, then this exception will be
                raised with the error message from SplatNet 3. This generally
                means that the query was successfully generated and the current
                tokens are still valid, but the query itself failed for some
                reason. This can happen if the user did not provide the correct
                required variables for the query, or if the variables provided
                were somehow invalid.

        Returns:
            QueryResponse: The response from the query. This object will contain
                the data from the query, and will also contain the query hash
                that was used to generate the query. This is useful for
                debugging purposes.
        """
        response = self.__query_hash(query_hash, variables)
        if response.status_code != 200:
            self.config.token_manager.generate_all_tokens()
            response = self.__query_hash(query_hash, variables)

        if "errors" in response.json():
            errors = response.json()["errors"]
            error_message = (
                "Query was successful but returned at least one error."
            )

            error_message += " Errors: " + json.dumps(errors, indent=4)
            raise SplatNetException(error_message)

        return QueryResponse(data=response.json()["data"])

    @retry(times=1, exceptions=ConnectionError)
    def query(self, query_name: str, variables: dict = {}) -> QueryResponse:
        """Queries Splatnet 3 and returns the data.

        This method will query SplatNet 3 and return the data from the query.
        This method will automatically retry once if the query fails and will
        regenerate the tokens if the query fails for whatever reason. Some
        queries require variables to be passed in, and this method will allow
        the user to pass in those variables. This method obtains the query hash
        from the ``auth`` module and automatically uses the appropriate query
        hash for the query name provided. As such, it is recommended to use this
        method rather than the ``query_hash`` method unless the user knows what
        they are doing.

        Args:
            query_name (str): The name of the query to use. This name must be
                valid, and is not case sensitive. For more information on the
                list of valid queries, see the documentation on the queries that
                are available.
            variables (dict): The variables to use in the query. Some queries do
                not require variables and so this argument can be omitted. If
                the query does require variables and this argument is omitted,
                then the query will fail and a ``SplatNetException`` will be
                raised with the error message from SplatNet 3. Defaults to {}.

        Raises:
            SplatNetException: If the query is successful but returns a JSON
                object with an ``errors`` key, then this exception will be
                raised with the error message from SplatNet 3. This generally
                means that the query was successfully generated and the current
                tokens are still valid, but the query itself failed for some
                reason. This can happen if the user did not provide the correct
                required variables for the query, or if the variables provided
                were somehow invalid.

        Returns:
            QueryResponse: The response from the query. This object will contain
                the data from the query, and will also contain the query hash
                that was used to generate the query. This is useful for
                debugging purposes.
        """
        response = self.__query(query_name, variables)
        if response.status_code != 200:
            self.config.token_manager.generate_all_tokens()
            response = self.__query(query_name, variables)

        if "errors" in response.json():
            errors = response.json()["errors"]
            error_message = (
                "Query was successful but returned at least one error."
            )

            error_message += " Errors: " + json.dumps(errors, indent=4)
            raise SplatNetException(error_message)

        return QueryResponse(data=response.json()["data"])



import os

from splatnet3_scraper.constants import ENV_VAR_NAMES, TOKENS


class EnvironmentVariablesManager:
    """Manages environment variables for the scraper.

    This class is used to manage the environment variables in an easy way. Any
    environment variable calls should be done through this class, it will handle
    obtaining the environment variables, allow additional environment variables
    to be added and defined, and will also handle the base environment variables
    that are required for this package to work.
    """

    BASE_TOKENS = [TOKENS.SESSION_TOKEN, TOKENS.GTOKEN, TOKENS.BULLET_TOKEN]

    def __init__(self) -> None:
        """Initializes the class and sets up the base environment variables."""
        self.variable_names = {}
        for token in self.BASE_TOKENS:
            self.variable_names[token] = ENV_VAR_NAMES[token]

    def token_to_variable(self, token: str) -> str:
        """Given the token name, returns the environment variable name.

        Args:
            token (str): The token name.

        Returns:
            str: The environment variable name.
        """
        return self.variable_names[token]

    def variable_to_token(self, variable: str) -> str:
        """Given the environment variable name, returns the token name.

        Args:
            variable (str): The environment variable name.

        Raises:
            KeyError: If the variable is not defined.

        Returns:
            str: The token name.
        """
        for token, variable_name in self.variable_names.items():
            if variable_name == variable:
                return token
        raise KeyError(f"Variable {variable} is not defined.")

    def add_token(self, token_name: str, variable_name: str) -> None:
        """Adds a new token to the environment variables.

        Args:
            token_name (str): The token name.
            variable_name (str): The environment variable name.
        """
        self.variable_names[token_name] = variable_name

    def remove_token(self, token_name: str) -> None:
        """Removes a token from the environment variables.

        Args:
            token_name (str): The token name.

        Raises:
            ValueError: If the token is a base token.
        """
        if token_name in self.BASE_TOKENS:
            raise ValueError(f"Cannot remove base token {token_name}.")
        del self.variable_names[token_name]

    def get(self, token: str) -> str | None:
        """Gets the environment variable for the given token.

        Args:
            token (str): The token to get the environment variable for.

        Returns:
            str | None: The environment variable, or None if it is not set.
        """
        return os.environ.get(self.token_to_variable(token))

    def get_all(self) -> dict[str, str | None]:
        """Gets all the environment variables.

        Returns:
            dict[str, str]: The environment variables.
        """
        return {token: self.get(token) for token in self.variable_names.keys()}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['splatnet3_scraper',
 'splatnet3_scraper.auth',
 'splatnet3_scraper.query',
 'splatnet3_scraper.scraper']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.28.1,<3.0.0']

extras_require = \
{'examples': ['pandas[examples]>=1.5.3,<2.0.0',
              'sqlalchemy[examples]>=2.0.1,<3.0.0',
              'psycopg2[examples]>=2.9.5,<3.0.0'],
 'parquet': ['pyarrow[parquet]>=10.0.1,<11.0.0']}

setup_kwargs = {
    'name': 'splatnet3-scraper',
    'version': '0.7.5',
    'description': 'Scraper for SplatNet 3 for Splatoon 3',
    'long_description': '# SplatNet 3 Scraper\n\n[![Tests Status](./reports/junit/tests-badge.svg?dummy=8484744)](https://htmlpreview.github.io/?https://github.com/cesaregarza/SplatNet3_Scraper/blob/main/reports/junit/report.html) ![Coverage Status](./reports/coverage/coverage-badge.svg?dummy=8484744) ![Flake8 Status](./reports/flake8/flake8-badge.svg?dummy=8484744) [![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n**SplatNet 3 Scraper** is a Python library for scraping data from the Splatoon 3 SplatNet 3 API. It is designed to be as lightweight as possible, with minimal dependencies to make it easy to integrate into other projects.\n\n**SplatNet 3 Scraper** started as a fork of **[s3s](https://github.com/frozenpandaman/s3s)**, but has since been rewritten from scratch while incorporating much of the login flow logic of s3s. As a result, I am deeply indebted to the authors of s3s for their work. This project would not have been possible without their efforts.\n\n## Table of Contents\n\n1. [Features](#features)\n2. [Documentation](#documentation)\n3. [Installation](#installation)\n4. [Usage](#usage)\n   - [Using the `scraper` module](#using-the-scraper-module)\n   - [Using the `query` module](#using-the-query-module)\n   - [Using the `auth` module](#using-the-auth-module)\n5. [Roadmap](#roadmap)\n6. [Contributing](#contributing)\n7. [License](#license)\n\n## Features\n\n- Lightweight and minimal dependencies. Only requires the `requests` library. Requires Python 3.10 or later.\n- The `scraper` module provides a user-level API that enables a quick and easy way to get data from the SplatNet 3 API, only requiring the user to provide their session token.\n- The `query` module provides a high-level API that provides a simple way to make queries to the SplatNet 3 API. It automatically handles authentication and query handling, and provides a simple interface for accessing the response data.\n- The `auth` module provides a low level API that allows for more fine-grained control over the scraping process. It greatly simplifies the process of authentication.\n- Compatibility with the configuration file format used by `s3s`.\n- Responses from the SplatNet 3 API can be saved and loaded from disk, currently supporting the following formats:\n  - JSON\n  - gzip-compressed JSON\n  - csv\n  - parquet (by installing `splatnet3_scraper[parquet]` or the `pyarrow` library)\n- Heavily documented codebase, with extensive docstrings and type annotations for nearly all functions and classes. The documentation is also available on [Read the Docs](https://splatnet3-scraper.readthedocs.io/en/latest/index.html).\n\n## Documentation\n\nDetailed documentation for SplatNet 3 Scraper, including usage instructions, examples, and API reference, is available on Read the Docs:\n\n[**SplatNet 3 Scraper Documentation**](https://splatnet3-scraper.readthedocs.io/en/latest/index.html)\n\nWe highly recommend referring to the documentation to get the most out of SplatNet 3 Scraper and understand its full capabilities.\n\n## Installation\n\n**SplatNet 3 Scraper** is currently under active development but is currently available on PyPI. It can be installed using pip:\n\n```bash\npip install splatnet3_scraper\n```\n\nNote that the current versions of **SplatNet 3 Scraper** are currently `v0.x.y`, which means that the API is not guaranteed to be stable and may change at any moment. As such, it is highly recommended that you pin the version of **SplatNet 3 Scraper** that you are using until the API is stabilized with the release of `v1.0.0`.\n\n## Usage\n\nThere are three ways to use **SplatNet 3 Scraper**. The first is to use the `scraper` module, which provides a top-level API that greatly simplifies the process of retrieving commonly requested data from SplatNet 3. This module is greatly recommended for most users. The second is to use the `query` module, which provides a high-level API that provides a simple interface to make queries to the SplatNet 3 API. This module is recommended for developers who can\'t find what they need with the `scraper` module. The third is to use the `auth` module, which provides a low-level API that gives the user the most control over the scraping process. This module is recommended for advanced developers who need to have full control over the authentication process. Whichever module you choose to use, all of them require providing a session token.\n\n### Using the `scraper` module\n\nThe `scraper` module is by far the easiest way to get data from the SplatNet 3 API and the module that is recommended for most users, especially those who are not highly experienced with Python. The `scraper` module provides multiple functions that can be used to retrieve commonly requested data from the SplatNet 3 API. The `scraper` module is designed to be used by users who are not highly experienced with Python or users who do not need to have full control over the scraping process.\n\nThis module is currently under active development and is not yet complete. Please check back later for more functions.\n\n### Using the `query` module\n\nThe `query` module is an easy-to-use module that enables fast and painless querying to the SplatNet 3 API. It handles authentication and query handling automagically, and provides a simple interface for accessing the response data. The `query` module is designed to be used by advanced users who need more control over the queries they make to the SplatNet 3 API. If you are looking for a simple way to get data from the SplatNet 3 API, you should use the `scraper` module instead.\n\nThe `query` module provides the `QueryHandler` class, which is used to make queries to the SplatNet 3 API. The `QueryHandler` class can be instantiated in one of a few ways: by providing a session token, by providing the path to a configuration file, or by loading environment variables.\n\n### Using the `auth` module\n\n:warning: **Warning: The `auth` module is intended for advanced users only. Most users should use the `scraper` or `query` modules for a simpler and more convenient experience.**\n\nThe `auth` module provides a low-level API that allows for more fine-grained control over the scraping process. It greatly simplifies the process of authentication and is designed for advanced developers who need full control over the authentication process.\n\nTo use the `auth` module, you will need to import the necessary components and handle the authentication flow manually. Please refer to the [documentation](https://splatnet3-scraper.readthedocs.io/en/latest/index.html) for detailed instructions and examples on how to use the `auth` module.\n\n#### Instantiating the `QueryHandler` class by providing a session token\n\n```python\nfrom splatnet3_scraper.query import QueryHandler\nhandler = QueryHandler.from_session_token("session_token")\nhandler.query("StageScheduleQuery")\n```\n\n#### Instantiating the `QueryHandler` class by providing the path to a configuration file\n\n```python\nfrom splatnet3_scraper.query import QueryHandler\nhandler = QueryHandler.from_config_file(".splatnet3_scraper")\nhandler.query("StageScheduleQuery")\n```\n\n#### Instantiating the `QueryHandler` class by loading environment variables\n\nThe following environment variables are supported:\n\n- `SN3S_SESSION_TOKEN`\n- `SN3S_GTOKEN`\n- `SN3S_BULLET_TOKEN`\n\n```python\nfrom splatnet3_scrape.query import QueryHandler\nhandler = QueryHandler.from_env()\nhandler.query("StageScheduleQuery")\n```\n\n#### Querying the SplatNet 3 API\n\nThe `QueryHandler` class provides a `query` method that can be used to make queries to the SplatNet 3 API. The `query` method takes a single argument, which is the name of the query to make. The `query` method returns a `QueryResponse` object, which contains the response data from the SplatNet 3 API. The `QueryResponse` object provides a `data` property that can be used to access the response data. The `QueryResponse` module also supports numpy-style indexing, which can be used to quickly and clearly access specific parts of the response data. For example, the following code will print the game mode name of the the current stage rotation schedule:\n\n```python\nfrom splatnet3_scraper.query import QueryHandler\nhandler = QueryHandler.from_env()\nresponse = handler.query("StageScheduleQuery")\nprint(response["xSchedules", "nodes", 0, "vsRule", "name"])\n```\n\n#### Saving and loading responses\n\nThe `QueryResponse` class provides a `parsed_json` method that can be used to generate a `JSONParser` object from the response data. The `JSONParser` class provides multiple ways of interacting with the given data, including the ability to save the data to disk in a variety of formats. There are currently four different formats that are supported and can be used by passing the desired format to a `to_*` method such as `to_json`. The following formats are supported:\n\n- JSON\n- gzip-compressed JSON\n- csv\n- parquet (by installing `splatnet3_scraper[parquet]` or the `pyarrow` library)\n\nNote: csv and parquet formats work by converting the response data from a nested dictionary to a columnar format. This is not recommended for single queries, but can be useful for interacting with large amounts of data as it deduplicates the JSON structure and allows for more efficient storage and querying.\n\nThe following code will save the response data to a file named `response.json` in the current directory:\n\n```python\nfrom splatnet3_scraper.query import QueryHandler\nhandler = QueryHandler.from_env()\nresponse = handler.query("StageScheduleQuery")\nresponse.parsed_json().to_json("response.json")\n```\n\nAdditionally, the `JSONParser` class provides a `from_*` method that can be used to load data from a file. The following code will load the response data from the file `response.json` in the current directory:\n\n```python\nfrom splatnet3_scraper.query import JSONParser\nparser = JSONParser.from_json("response.json")\n```\n\n## Symbols\n\n| Symbol | Meaning |\n| ------ | ------- |\n| :white_check_mark: | Implemented |\n| :construction: | In progress |\n| :world_map: | Planned |\n| :x: | Not planned |\n\n## Roadmap\n\n| Feature | Status |\n| ------- | ------ |\n| Support for the SplatNet 3 API | :white_check_mark: |\n| Full support for the SplatNet 3 API | :white_check_mark: |\n| Support for the SplatNet 2 API | :x: |\n| Obtaining session tokens | :white_check_mark: |\n| Full documentation | :white_check_mark: |\n| Full unit test coverage | :white_check_mark: |\n| Columnar data format support | :construction: |\n| CLI interface | :x: |\n| Integration with stat.ink | :x: |\n| PyPI package | :white_check_mark: |\n| Docker image | :world_map: |\n| Executable binary | :x: |\n\n## Contributing\n\nWe welcome contributions to SplatNet 3 Scraper! For detailed information on how to contribute, please refer to our [CONTRIBUTING.md](./CONTRIBUTING.md) file.\n\nTo report issues or request new features, please open an issue on the GitHub repository.\n\n## License\n\nSplatNet 3 Scraper is licensed under the GPLv3. See the [LICENSE](./LICENSE) file for more details.\n',
    'author': 'Cesar E Garza',
    'author_email': 'cesar@cegarza.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



SPLATNET_URL = "https://api.lp1.av5ja.srv.nintendo.net"
GRAPHQL_URL = SPLATNET_URL + "/api/graphql"
GRAPH_QL_REFERENCE_URL = (
    "https://raw.githubusercontent.com"
    "/imink-app/SplatNet3/master/Data/splatnet3_webview_data.json"
)
IOS_APP_URL = (
    "https://apps.apple.com/us/app/nintendo-switch-online/id1234806557"
)
IMINK_URL = "https://api.imink.app/f"
DEFAULT_USER_AGENT = (
    "Mozilla/5.0 (Linux; Android 11; Pixel 5) "
    + "AppleWebKit/537.36 (KHTML, like Gecko) "
    + "Chrome/94.0.4606.61 "
    + "Mobile Safari/537.36"
)
WEB_VIEW_VERSION_FALLBACK = "3.0.0-0742bda0"
APP_VERSION_FALLBACK = "2.5.0"
HASHES_FALLBACK = {
    "BankaraBattleHistoriesQuery": "0438ea6978ae8bd77c5d1250f4f84803",
    "BankaraBattleHistoriesRefetchQuery": "92b56403c0d9b1e63566ec98fef52eb3",
    "BattleHistoryCurrentPlayerQuery": "49dd00428fb8e9b4dde62f585c8de1e0",
    "CatalogQuery": "ff12098bad4989a813201b00ff22ac4e",
    "CatalogRefetchQuery": "60a6592c6ee8e47245020ae0d314d378",
    "ChallengeQuery": "8a079214500148bf88a8fce1d7209b90",
    "ChallengeRefetchQuery": "34aedc79f96b8613501bba465295f779",
    "CheckinQuery": "5d0d1b45ebf4e324d0dae017d9df06d2",
    "CheckinWithQRCodeMutation": "daffd9621680664dbf19d27e87484ac7",
    "ConfigureAnalyticsQuery": "f8ae00773cc412a50dd41a6d9a159ddd",
    "CoopHistoryDetailQuery": "379f0d9b78b531be53044bcac031b34b",
    "CoopHistoryDetailRefetchQuery": "d3188df2fd4436870936b109675e2849",
    "CoopHistoryQuery": "91b917becd2fa415890f5b47e15ffb15",
    "CoopPagerLatestCoopQuery": "eb947416660e0a7520549f6b9a8ffcd7",
    "CoopRecordBigRunRecordContainerPaginationQuery": (
        "2b83817b6e88b202d25939fe04658d33"
    ),
    "CoopRecordQuery": "b2f05c682ed2aeb669a86a3265ceb713",
    "CoopRecordRefetchQuery": "15035e6c4308b32d1a77e87398be5cd4",
    "CreateMyOutfitMutation": "31ff008ea218ffbe11d958a52c6f959f",
    "DetailFestRecordDetailQuery": "96c3a7fd484b8d3be08e0a3c99eb2a3d",
    "DetailFestRefethQuery": "18c7c465b18de5829347b7a7f1e571a1",
    "DetailFestVotingStatusRefethQuery": "92f51ed1ab462bbf1ab64cad49d36f79",
    "DetailRankingQuery": "cc38f388c51f9930bd7cca966893f1b4",
    "DetailTabViewWeaponTopsArRefetchQuery": "a6782a0c692e8076656f9b4ab613fd82",
    "DetailTabViewWeaponTopsClRefetchQuery": "8d3c5bb2e82d6eb32a37eefb0e1f8f69",
    "DetailTabViewWeaponTopsGlRefetchQuery": "b23468857c049c2f0684797e45fabac1",
    "DetailTabViewWeaponTopsLfRefetchQuery": "d46f88c2ea5c4daeb5fe9d5813d07a99",
    "DetailTabViewXRankingArRefetchQuery": "6de3895bd90b5fa5220b5e9355981e16",
    "DetailTabViewXRankingClRefetchQuery": "3ab25d7f475cb3d5daf16f835a23411b",
    "DetailTabViewXRankingGlRefetchQuery": "d62ec65b297968b659103d8dc95d014d",
    "DetailTabViewXRankingLfRefetchQuery": "d96057b8f46e5f7f213a35c8ea2b8fdc",
    "DetailVotingStatusQuery": "53ee6b6e2acc3859bf42454266d671fc",
    "DownloadSearchReplayQuery": "8e904b52b5080b6f4b4448a50762362c",
    "FestRecordQuery": "44c76790b68ca0f3da87f2a3452de986",
    "FestRecordRefetchQuery": "73b9837d0e4dd29bfa2f1a7d7ee0814a",
    "FriendListQuery": "f0a8ebc384cf5fbac01e8085fbd7c898",
    "FriendListRefetchQuery": "aa2c979ad21a1100170ddf6afea3e2db",
    "GesotownQuery": "a43dd44899a09013bcfd29b4b13314ff",
    "GesotownRefetchQuery": "951cab295eafdbeccfc2e718d7a98646",
    "HeroHistoryQuery": "fbee1a882371d4e3becec345636d7d1c",
    "HeroHistoryRefetchQuery": "4f9ae2b8f1d209a5f20302111b28f975",
    "HistoryRecordQuery": "f09da9d24d888797fdfb2f060dbdf4ed",
    "HistoryRecordRefetchQuery": "d997d8e3875d50d3a1dc7e8a756e9e07",
    "HomeQuery": "22e2fa8294168003c21b00c333c35384",
    "JourneyChallengeDetailQuery": "5a199948d059985bd758cc0175131f4a",
    "JourneyChallengeDetailRefetchQuery": "e7414c7a64bf80bb50ce21d5ccfde772",
    "JourneyQuery": "bc71fc0264f3f72256724b069f7a4097",
    "JourneyRefetchQuery": "09eee118fa16415d6bc3846bc6e5d8e5",
    "LatestBattleHistoriesQuery": "0176a47218d830ee447e10af4a287b3f",
    "LatestBattleHistoriesRefetchQuery": "7161210aad0793e58e76f20e0443855e",
    "MyOutfitDetailQuery": "d935d9e9ba7a5b6b5d6ece7f253304fc",
    "MyOutfitsQuery": "81d9a6849467d2aa6b1603ebcedbddbe",
    "MyOutfitsRefetchQuery": "10db4e349f3123c56df14e3adec2ee6f",
    "PagerLatestVsDetailQuery": "0329c535a32f914fd44251be1f489e24",
    "PagerUpdateBattleHistoriesByVsModeQuery": (
        "094a9b44ff21e8c409d6046fc1af9dfe"
    ),
    "PhotoAlbumQuery": "7e950e4f69a5f50013bba8a8fb6a3807",
    "PhotoAlbumRefetchQuery": "53fb0ad32c13dd9a6e617b1158cc2d41",
    "PrivateBattleHistoriesQuery": "8e5ae78b194264a6c230e262d069bd28",
    "PrivateBattleHistoriesRefetchQuery": "89bc61012dcf170d9253f406ebebee67",
    "RankingHoldersFestTeamRankingHoldersPaginationQuery": (
        "f488fccdad37b9e19aed50a8d6e83a24"
    ),
    "RegularBattleHistoriesQuery": "3baef04b095ad8975ea679d722bc17de",
    "RegularBattleHistoriesRefetchQuery": "4c95233c8d55e7c8cc23aae06109a2e8",
    "ReplayModalReserveReplayDownloadMutation": (
        "87bff2b854168b496c2da8c0e7f3e5bc"
    ),
    "ReplayQuery": "e9cbaa835977b6c6de77ca7a4be15b24",
    "ReplayUploadedReplayListRefetchQuery": "3bd200163e63bfff42ab60a244cac042",
    "SaleGearDetailOrderGesotownGearMutation": (
        "b79b7a101a243912754f72437e2ad7e5"
    ),
    "SaleGearDetailQuery": "6eb1b255b2cf04c08041567148c883ad",
    "SettingQuery": "73bd677ed986ad2cb7004ceabfff4d38",
    "StageRecordQuery": "f08a932d533845dde86e674e03bbb7d3",
    "StageRecordsRefetchQuery": "2fb1b3fa2d40c9b5953ea1ae263e54c1",
    "StageScheduleQuery": "011e394c0e384d77a0701474c8c11a20",
    "SupportButton_SupportChallengeMutation": (
        "991bace9e8c52d63084cd1570a97a5b4"
    ),
    "UpdateMyOutfitMutation": "bb809066282e7d659d3b9e9d4e46b43b",
    "VotesUpdateFestVoteMutation": "a2c742c840718f37488e0394cd6e1e08",
    "VsHistoryDetailPagerRefetchQuery": "994cf141e55213e6923426caf37a1934",
    "VsHistoryDetailQuery": "291295ad311b99a6288fc95a5c4cb2d2",
    "WeaponRecordQuery": "5f279779e7081f2d14ae1ddca0db2b6e",
    "WeaponRecordsRefetchQuery": "6961f618fcef440c81509b205465eeec",
    "XBattleHistoriesQuery": "6796e3cd5dc3ebd51864dc709d899fc5",
    "XBattleHistoriesRefetchQuery": "94711fc9f95dd78fc640909f02d09215",
    "XRankingDetailQuery": "d5e4924c05891208466fcba260d682e7",
    "XRankingDetailRefetchQuery": "fb960404299958248b3c0a2fbb444c35",
    "XRankingQuery": "d771444f2584d938db8d10055599011d",
    "XRankingRefetchQuery": "5149402597bd2531b4eea04692d8bfd5",
    "myOutfitCommonDataEquipmentsQuery": "d29cd0c2b5e6bac90dd5b817914832f8",
    "myOutfitCommonDataFilteringConditionQuery": (
        "d02ab22c9dccc440076055c8baa0fa7a"
    ),
    "refetchableCoopHistory_coopResultQuery": (
        "50be9b694c7c6b99b7a383e494ec5258"
    ),
    "useCurrentFestQuery": "c0429fd738d829445e994d3370999764",
    "useShareMyOutfitQuery": "3ba5572efce5bebbd859fc2d269d223c",
}


class TOKENS:
    SESSION_TOKEN = "session_token"
    GTOKEN = "gtoken"
    BULLET_TOKEN = "bullet_token"


TOKEN_EXPIRATIONS = {
    TOKENS.GTOKEN: (60 * 60 * 6) + (60 * 30),
    TOKENS.BULLET_TOKEN: (60 * 60 * 2),
}
ENV_VAR_NAMES = {
    TOKENS.SESSION_TOKEN: "SN3S_SESSION_TOKEN",
    TOKENS.GTOKEN: "SN3S_GTOKEN",
    TOKENS.BULLET_TOKEN: "SN3S_BULLET_TOKEN",
}

PRIMARY_ONLY = [
    "Comeback",
    "Last-Ditch Effort",
    "Opening Gambit",
    "Tenacity",
    "Ability Doubler",
    "Haunt",
    "Ninja Squid",
    "Respawn Punisher",
    "Thermal Ink",
    "Drop Roller",
    "Object Shredder",
    "Stealth Jump",
]

ABILITIES = [
    "Ink Saver (Main)",
    "Ink Saver (Sub)",
    "Ink Recovery Up",
    "Run Speed Up",
    "Swim Speed Up",
    "Special Charge Up",
    "Special Saver",
    "Special Power Up",
    "Quick Respawn",
    "Quick Super Jump",
    "Sub Power Up",
    "Ink Resistance Up",
    "Sub Resistance Up",
    "Intensify Action",
]

ALL_ABILITIES = PRIMARY_ONLY + ABILITIES
